# PAX - the Pythonic API for XML
# developed by Lalo Martins <lalo@laranja.org>
# this file is placed on public domain

from backwards_compatibility import *
from utils import XML, pack_list
from types import StringType, UnicodeType, TupleType, ListType
from sys import getdefaultencoding
_replace='replace'

# PAX containers - an Element or, for example, a Document
class Container(object):
    def __init__(self):
        self.children = []

    def __getitem__(self, index):
        return self.children[index]

    def __iter__(self):
        return iter(self.children)

    def __len__(self):
        return len(self.children)

    def __setitem__(self, index, value):
        self.children[index] = value

    def append(self, child):
        self.children.append(child)

    def pack(self):
        "remove Nones, expand tuples"
        self.children = pack_list(self.children)

    def copy(self):
        copy = self.__class__()
        # children is shallow-copied!
        copy.children = self.children[:]
        return copy

def usafe(obj):
    "check if a string is unicode, if not, safely convert it"
    if not obj or type(obj) is UnicodeType:
        return obj
    if not type(obj) is StringType:
        obj = str(obj)
    return unicode(obj, getdefaultencoding(), _replace)
    return obj

# The PAX object
class Element(Container):
    __core_slots = ('ns', 'name', 'attributes', 'children',
                    'nsdecls', 'lame_namespaces', 'raw')
    def __init__(self, ns=u'', name=u'', attributes={}, nsdecls={}, raw=u''):
        self.ns = ns
        self.name = name
        self.attributes = attributes
        self.children = []
        self.nsdecls = nsdecls
        self.lame_namespaces = {}
        for dprefix, duri in self.nsdecls.items():
            if not attributes.has_key(duri):
                attributes[duri] = {}
        self.raw = raw

    def copy(self):
        copy = self.__class__(self.ns, self.name, {}, self.nsdecls.copy(), self.raw)
        # attributes is copied with one level of depth
        for ns, attrs in self.attributes.items():
            copy.attributes[ns] = attrs.copy()
        # children is shallow-copied!
        copy.children = self.children[:]
        copy.lame_namespaces = self.lame_namespaces.copy()
        # non-standard attributes are also shallow-copied - ATTENTION
        for name, value in self.__dict__.items():
            if name not in self.__core_slots:
                setattr(copy, name, value)
        return copy

    def __repr__(self):
        return '<PAX Element: %s>' % self.__dict__

    def __self_xml__(self, ns_map):
        ns_map = ns_map.copy()
        for prefix, uri in self.nsdecls.items():
            ns_map[uri] = prefix
        for prefix, id in self.lame_namespaces.items():
            ns_map[id] = prefix
        prefix = ns_map.get(self.ns, u'')
        if prefix is None: prefix = u''
        if prefix: prefix += u':'
        result = u'<' + prefix + self.name
        for dprefix, duri in self.nsdecls.items():
            result += ' xmlns'
            if dprefix:
                result += ':' + dprefix
            result += '="' + duri + '"'
        for ans, attrs in self.attributes.items():
            result += self.__attribute_ns_xml__(ans, attrs, ns_map)
        return result, prefix

    def __attribute_ns_xml__(self, ns, attrs, ns_map):
        if ns == self.ns:
            prefix = u''
        else:
            prefix = usafe(ns_map.get(ns, u''))
            if prefix is None: prefix = u''
            if prefix: prefix += u':'
        result = u''
        for name, value in attrs.items():
            result += self.__attribute_xml__(name, value, prefix)
        return result

    def __attribute_xml__(self, name, value, prefix):
        if value is None: value = u''
        name = usafe(name)
        value = usafe(value)
        if name:
            return u' ' + prefix + name + u'="' + value + u'"'
        return u''

    def __xml__(self, ns_map):
        if ns_map is None:
            result = self.raw[:-1]
            if result[-1] == '/':
                result = result[:-1]
            prefix = result[1:result.find(self.name)]
        else:
            result, prefix = self.__self_xml__(ns_map)
        result += self.__children_xml__(prefix, ns_map)
        return result

    def __children_xml__(self, prefix, ns_map):
        "render children to XML; can be overridden to avoid self-closing, for example"
        if self.children:
            return u'>' + XML(self.children, ns_map) + u'</' + prefix + self.name + u'>'
        else:
            return u' />'

    def __nonzero__(self):
        return 1
            

    # todo: __str__

# character reference
class CharRef(object):
    def __init__(self, num=0):
        self.num = num

    def __repr__(self):
        return '<PAX Character Reference: %s (%s)>' % (self.num, self.unichr(self.num))

    def __xml__(self, ns_map):
        #FIXME: what is the expected format?
        return u'&#%s;' % self.num

# entity reference
class EntityRef(object):
    def __init__(self, name=u''):
        self.name = name

    def __repr__(self):
        return '<PAX Entity Reference: %s>' % self.name

    def __xml__(self, ns_map):
        return u'&%s;' % self.name

# comment
class Comment(object):
    def __init__(self, text=u''):
        self.text = text

    def __repr__(self):
        return '<PAX Comment: %s>' % self.text

    def __xml__(self, ns_map):
        return u'<!--%s-->' % self.text

# xml/sgml declaration
class Declaration(object):
    def __init__(self, text=u''):
        self.text = text

    def __repr__(self):
        return '<PAX Declaration: %s>' % self.text

    def __xml__(self, ns_map):
        return u'<!%s>' % self.text
