# PAX - the Pythonic API for XML
# developed by Lalo Martins <lalo@laranja.org>
# this file is placed on public domain

from backwards_compatibility import *
from utils import pack_list
import types

class StopTransform(Exception):
    def __init__(self, result=None, *args):
        Exception.__init__(self, *args)
        self.result = (result,)

class Engine(object):
    def __init__(self):
        self.handlers = ()

    def add_handler(self, handler, optional=None):
        if optional is None:
            if hasattr(handler, 'ns'):
                ns = handler.ns
            else:
                raise TypeError, 'either pass in a namespace uri, or a handler with a "ns" attribute'
        else:
            ns = handler
            handler = optional
        self.handlers += ((ns, hash(ns), handler),)

    def initialize(self, context):
        context.tr_engine = self
        context.postprocess_queue = []
        for ns, h, handler in self.handlers:
            try:
                handler.initialize(context)
            except (AttributeError, TypeError):
                pass

    def transform(self, element, context):
        try:
            result = element.copy()
        except AttributeError:
            return element
        try:
            __traceback_info__ = element.raw
        except AttributeError:
            pass
        processed = []
        context.postprocess_queue.append(processed)
        try:
            e_ns_h = hash(element.ns)
        except AttributeError:
            e_ns_h = None
        try:
            for ns, h, handler in self.handlers:
                if not hasattr(result, 'attributes'):
                    break
                # We compare the hash of the ns against the hash of the element's ns
                # this avoids an expensive unicode character-by-character comparison,
                # in exchange for hash computation which is cached.
                #
                # However, this also means you can't change an element's ns in one
                # transform handler and expect it to affect future handlers.
                #
                # Then again, if you wanted to do that, you probably should be shot.
                #
                #
                #
                #
                # (just kidding)
                if result.attributes.has_key(ns) or e_ns_h == h:
                    processed.append(handler)
                    result = handler(result, context)
        except StopTransform, data:
            result = data.result
            if type(result) is types.ListType:
                # further mark it as "immutable"
                result = tuple(result)
        else:
            if not hasattr(result, 'join'):
                for index, child in enumerate(result):
                    result[index] = self.transform(child, context)
        try:
            result.pack()
        except AttributeError:
            try:
                result = pack_list(result)
            except:
                pass
        try:
            result = self.postprocess(result, context)
        except StopTransform, data:
            result = data.result
        context.postprocess_queue.pop()
        return result

    def postprocess(self, element, context):
        processed = context.postprocess_queue[-1]
        while processed:
            handler = processed.pop(0)
            try:
                pp = handler.postprocess
            except AttributeError:
                continue
            if not callable(pp): continue
            element = pp(element, context)
        return element


class AttributeHandler(object):
    "Handler that handles each attribute in turn"
    def __init__(self, name=u'', ns=u''):
        self.name = name
        self.ns = ns
        self.handlers = ()
        self.preprocess = None
        self.postprocess = None
        self.initialize = None

    def add_handler(self, name, handler):
        self.handlers += ((name, handler),)

    def __call__(self, element, context):
        result = element
        if self.preprocess is not None:
            result = self.preprocess(result, context)
        for name, handler in self.handlers:
            try:
                value = result.attributes[self.ns][name]
            except KeyError:
                continue
            result = handler(result, context, value)
        return result

compiled_suffix = u'_compiled_data'

class CompiledInformationHandler(object):
    "Handler that handles compiled information"
    def __init__(self, name=u'', ns=u''):
        self.name = name
        self.ns = ns
        self.handlers = {}
        self.compiled_name = name + compiled_suffix
        self.preprocess = None
        self.postprocess = None
        self.initialize = None

    def add_handler(self, name, handler):
        self.handlers[name] = handler

    def __call__(self, element, context):
        result = element
        # make a copy of the compiled data, because Element.copy()
        # will have shallow-copied it
        compiled = getattr(result, self.compiled_name)[:]
        setattr(result, self.compiled_name, compiled)
        if self.preprocess is not None:
            result = self.preprocess(result, context)
        for name, value in compiled:
            __traceback_info__ = (name, value)
            result = self.handlers[name](result, context, value)
        return result

class AttributeCompiler(AttributeHandler):
    "Handler that compiles atribute contents"
    def __init__(self, name=u'', ns=u''):
        AttributeHandler.__init__(self, name, ns)
        self.compiled_name = name + compiled_suffix

    def __call__(self, element, context):
        compiled = []
        setattr(element, self.compiled_name, compiled)
        if self.preprocess is not None:
            self.preprocess(element, context)
        for name, handler in self.handlers:
            try:
                value = element.attributes[self.ns][name]
            except KeyError:
                continue
            compiled.append((name, handler(element, context, value)))
        return element

try:
    import psyco
    psyco.bind(Engine)
    psyco.bind(AttributeHandler)
    psyco.bind(CompiledInformationHandler)
    psyco.bind(AttributeCompiler)
except:
    pass
