# PAX - the Pythonic API for XML
# developed by Lalo Martins <lalo@laranja.org>
# this file is placed on public domain

# parse text with embedded xmlish tags,
# or lax (not really conformant) xmlish documents

# to parse real html you'll want to use htmlpaxparser
# which copes with broken browsers for stuff like
# <textarea />, <a />, <select /> etc

import core
from HTMLParser import HTMLParser
from backwards_compatibility import *

class Document(core.Container):
    def __repr__(self):
        return u'<PAX Document: %s>' % self.children

# get a namespace, if provided
def split_ns(spec, default=u''):
    if ':' in spec:
        return spec.split(':', 1)
    else:
        return [default, spec]

class _parse_in_progress(HTMLParser):
    def __init__(self):
        self.result = Document()
        self.stack = [self.result]
        self.namespaces = []
        self.lame_namespaces = {}
        self.ns_pushed = False
        self.ns_popped = False
        HTMLParser.__init__(self)

    def __raw(self):
        return self.get_starttag_text()

    def ns(self, prefix):
        for layer in self.namespaces:
            if layer.has_key(prefix):
                return layer[prefix]
        return self.lame_namespaces.setdefault(prefix, ('lame_ns', prefix))

    def handle_starttag(self, rawname, raw_attributes):
        ns, name = split_ns(rawname)
        attributes = {}
        for aspec, avalue in raw_attributes:
            if aspec == u'xmlns':
                self.add_namespace(None, avalue)
            if aspec[:6] == u'xmlns:':
                self.add_namespace(aspec[6:], avalue)
        for aspec, avalue in raw_attributes:
            if aspec[:5] == u'xmlns':
                continue
            ans, aname = split_ns(aspec, ns)
            ans = self.ns(ans)
            attributes.setdefault(ans, {})[aname] = avalue
        if self.ns_pushed:
            nsdecls = self.namespaces[0]
        else:
            nsdecls = {}
        ns = self.ns(ns)
        element = core.Element(ns, name, attributes, nsdecls, self.__raw())
        element.rawname = rawname
        self.stack.append(element)
        self.ns_pushed = False

    def handle_endtag(self, name):
        topname = getattr(self.stack[-1], 'rawname', None)
        if len(self.stack) > 1:
            self.stack[-2].append(self.stack[-1])
            self.stack.pop()
        self.ns_popped = False
        if name != topname and len(self.stack) > 1:
            #print "minimizing:", name, topname
            self.handle_endtag(name)

    def close(self):
        while len(self.stack) > 1:
            self.stack[-2].append(self.stack[-1])
            self.stack.pop()
        HTMLParser.close(self)
        for element in self.result:
            if isinstance(element, core.Element):
                element.lame_namespaces = self.lame_namespaces

    def handle_decl(self, data):
        self.stack[-1].append(core.Declaration(data))

    def handle_data(self, data):
        self.stack[-1].append(data)

    def handle_charref(self, data):
        self.stack[-1].append(core.CharRef(int(data)))

    def handle_entityref(self, data):
        self.stack[-1].append(core.EntityRef(data))

    def add_namespace(self, prefix, url):
        if self.ns_pushed:
            d = self.namespaces[0]
        else:
            self.ns_pushed = True
            d = {}
            self.namespaces.insert(0, d)
        d[prefix] = url

    def pop_namespace(self, prefix):
        if not self.ns_popped:
            del self.namespaces[0]
        self.ns_popped = True
    
    def handle_comment(self, data):
        self.stack[-1].append(core.Comment(data))

def parse(data, builtin_namespaces=None, encoding=None):
    if encoding is None:
        # FIXME: can we get the encoding from the HTML?
        encoding = 'latin1'
    if type(data) is type(''):
        data = unicode(data, encoding)
    parser = _parse_in_progress()
    if builtin_namespaces is not None:
        parser.namespaces.append(builtin_namespaces)
    parser.feed(data)
    parser.close()
    return parser.result
